@@ -17,8 +17,9 @@ from typing import Iterable, Optional, Tuple, Union
 
 from lb.nightly.configuration import Slot
 
+from .utils import LockTakenError
+
 MAX_TIME_BETWEEN_CONNECTIONS = datetime.timedelta(minutes=5)
-SECONDS_BETWEEN_LOCK_ATTEMPTS = 10
 
 
 class Database:
@@ -303,6 +304,7 @@ class Database:
     def checkout_complete(
         self,
         project,
+        artifact,
         report,
         worker_task_id,
     ):
@@ -320,7 +322,7 @@ class Database:
             proj["tree"] = report.tree
             proj["warnings"] = []
             proj["errors"] = []
-            proj["artifact"] = project.artifacts("checkout")
+            proj["artifact"] = artifact
             try:
                 proj["dependencies"] = project.dependencies()
             except AttributeError:
@@ -334,6 +336,35 @@ class Database:
 
         self.apply(update, self[project.slot])
 
+    def reuse_checkout(self, project, artifact):
+        """
+        Set checkout summary details from another checkout that produced the
+        same artifact.
+
+        Return True if such summary could be found, False otherwise.
+        """
+        summary = self.get_artifact_summary(artifact)
+
+        if summary:
+
+            def update_summary(doc):
+                if "checkout" not in doc:
+                    doc["checkout"] = {}
+                if "projects" not in doc["checkout"]:
+                    doc["checkout"]["projects"] = {}
+                doc["checkout"]["projects"][project.name] = summary
+                for prj in doc["config"]["projects"]:
+                    if prj["name"] == project.name:
+                        prj["dependencies"] = summary["dependencies"]
+                        break
+
+            self.apply(update_summary, self[project.slot])
+
+            return True
+
+        else:
+            return False
+
     def build_start(self, project, platform, worker_task_id):
         def update(doc):
             if "builds" not in doc:
@@ -359,7 +390,7 @@ class Database:
 
         self.apply(update, self[project.slot])
 
-    def build_complete(self, project, platform, reports, worker_task_id):
+    def build_complete(self, project, platform, artifact, reports, worker_task_id):
         def update(doc):
             proj = doc["builds"][platform][project.name]
             assert (
@@ -372,7 +403,7 @@ class Database:
             proj["warnings"] = reports["warnings"]
             proj["errors"] = reports["errors"]
             proj["retcode"] = reports["retcode"]
-            proj["artifact"] = project.artifacts("build", platform)
+            proj["artifact"] = artifact
 
         self.apply(update, self[project.slot])
 
@@ -401,7 +432,7 @@ class Database:
 
         self.apply(update, self[project.slot])
 
-    def tests_complete(self, project, platform, reports, worker_task_id):
+    def tests_complete(self, project, platform, artifact, reports, worker_task_id):
         def update(doc):
             proj = doc["tests"][platform][project.name]
             assert (
@@ -411,7 +442,7 @@ class Database:
                 proj["worker_task_id"] == worker_task_id
             ), "trying to complete a test started by somebody else"
             proj["completed"] = str(datetime.datetime.now())
-            proj["artifact"] = project.artifacts("test", platform)
+            proj["artifact"] = artifact
             proj["results"] = reports["results"]
 
         self.apply(update, self[project.slot])
@@ -436,29 +467,61 @@ class Database:
         """
         CouchDB backed resource lock.
 
-        The context manager blocks if the lock "id" is already taken.
+        The context manager raises LockTakenError if the lock "id" is already taken.
         """
         from cloudant.error import CloudantDatabaseException
 
         doc_id = f"lock:{id}"
-        while True:
-            try:
-                self._db.create_document(
-                    {
-                        "_id": doc_id,
-                        "type": "lock",
-                        "acquired": str(datetime.datetime.now()),
-                        "info": info,
-                    },
-                    throw_on_exists=True,
-                )
-                break
-            except CloudantDatabaseException:
-                # somebody already holds the lock, we try again a bit later
-                sleep(SECONDS_BETWEEN_LOCK_ATTEMPTS)
+        try:
+            self._db.create_document(
+                {
+                    "_id": doc_id,
+                    "type": "lock",
+                    "acquired": str(datetime.datetime.now()),
+                    "info": info,
+                },
+                throw_on_exists=True,
+            )
+        except CloudantDatabaseException:
+            # somebody already holds the lock
+            raise LockTakenError(doc_id)
 
         try:
             yield doc_id
 
         finally:
             self._db[doc_id].delete()
+
+    def reuse_artifact(self, project, artifact, platform=None, stage=None):
+        """
+        Set summary details from another checkout/build/test that produced the
+        same artifact.
+
+        Return True if such summary could be found, False otherwise.
+        """
+        doc_name = f"{stage}s" if stage else "checkout"
+        subdoc_name = platform or "projects"
+
+        summary = self.get_artifact_summary(artifact)
+
+        if summary:
+
+            def update_summary(doc):
+                if doc_name not in doc:
+                    doc[doc_name] = {}
+                if subdoc_name not in doc[doc_name]:
+                    doc[doc_name][subdoc_name] = {}
+                doc[doc_name][subdoc_name][project.name] = summary
+                if not platform and not stage:
+                    # need to update dependencies in case of updating checkout
+                    for prj in doc["config"]["projects"]:
+                        if prj["name"] == project.name:
+                            prj["dependencies"] = summary["dependencies"]
+                            break
+
+            self.apply(update_summary, self[project.slot])
+
+            return True
+
+        else:
+            return False
