@@ -186,12 +186,37 @@ VIEWS = [
         }
       }
     }
+    if (doc.tests) {
+      for(var platform in doc.tests) {
+        for(var project in doc.tests[platform]) {
+            var info = doc.tests[platform][project];
+            if (info.artifact) {
+              emit(info.artifact, info);
+            }
+        }
+      }
+    }
   }
 }
 """,
             },
         },
     },
+    {
+        "_id": "_design/state",
+        "language": "javascript",
+        "options": {"partitioned": False},
+        "views": {
+            "aborted": {
+                "map": """function (doc) {
+  if (doc.type == "slot-info" && doc.aborted) {
+    emit(doc.config.metadata.flavour + "/" + doc.slot + "/" + doc.build_id, doc.aborted);
+  }
+}
+"""
+            }
+        },
+    },
     {
         "_id": "_design/mango_indexes",
         "language": "query",
@@ -246,3 +271,8 @@ def init_db(db_url=None, overwrite=False):
     security["members"] = {}
     security.save()
     return db
+
+
+class LockTakenError(Exception):
+    def __str__(self):
+        return "somebody already holds the lock for {}".format(*self.args)
